"""
Changelog and version history for xautic.

.. versionadded:: 1.0.0
    Built and setup the structure for the application.

.. versionadded:: 1.0.0
    Support for `debug` decorator was added along with major code
    refactoring.
"""

__all__ = ["__version__"]

__version__ = "1.0.0"
